import unittest
from unittest import mock

from quizler.main import create_parser, main
from tests.utils import mock_envs, mock_argv


class TestCreateParser(unittest.TestCase):

    def setUp(self):
        self.parser = create_parser()

    def test_with_empty_args(self):
        with self.assertRaises(SystemExit):
            self.parser.parse_args([])

    def test_unknown_command(self):
        with self.assertRaises(SystemExit):
            self.parser.parse_args(['ping'])

    def test_common_command(self):
        args = self.parser.parse_args(['common'])
        self.assertEqual(args.command, 'common')

    def test_sets_command(self):
        args = self.parser.parse_args(['sets'])
        self.assertEqual(args.command, 'sets')

    def test_apply_command_without_args(self):
        with self.assertRaises(SystemExit):
            self.parser.parse_args(['apply'])

    def test_apply_command_with_all_args(self):
        args = self.parser.parse_args(['apply', 'pattern', 'repl', 'set_name'])
        self.assertEqual(args.command, 'apply')
        self.assertEqual(args.pattern, 'pattern')
        self.assertEqual(args.repl, 'repl')
        self.assertEqual(args.set_name, 'set_name')


@mock_envs(CLIENT_ID='client_id', USER_ID='user_id')
class TestMain(unittest.TestCase):

    @mock.patch('quizler.main.get_common_terms')
    @mock.patch('quizler.main.print_common_terms')
    @mock_argv('common')
    def test_common(self, mock_get_common_terms, mock_print_common_terms):
        main()
        mock_get_common_terms.assert_called_once()
        mock_print_common_terms.assert_called_once()

    @mock.patch('quizler.main.get_user_sets')
    @mock_argv('sets')
    def test_sets(self, mock_get_user_sets):
        main()
        mock_get_user_sets.assert_called_once()

    @mock.patch('quizler.main.apply_regex')
    @mock_argv('apply', 'pattern', 'repl', 'set_name')
    def test_apply(self, mock_apply_regex):
        main()
        mock_apply_regex.assert_called_once()



import unittest
from unittest import mock

from quizler.lib import api_call, get_api_envs
from tests.utils import mock_envs


class TestGetApiEnvs(unittest.TestCase):

    @mock_envs()
    def test_no_envs(self):
        with self.assertRaises(ValueError):
            get_api_envs()

    @mock_envs(USER_ID='user_id')
    def test_no_client_id(self):
        with self.assertRaises(ValueError):
            get_api_envs()

    @mock_envs(CLIENT_ID='client_id')
    def test_no_user_id(self):
        with self.assertRaises(ValueError):
            get_api_envs()

    @mock_envs(CLIENT_ID='client_id', USER_ID='user_id')
    def test_all_envs_are_set(self):
        client_id, user_id = get_api_envs()
        self.assertEqual(client_id, 'client_id')
        self.assertEqual(user_id, 'user_id')


class TestApiCall(unittest.TestCase):

    @mock.patch('requests.get')
    def test_unknown_endpoint(self, mock_get):
        mock_get.return_value.status_code = 404
        with self.assertRaises(ValueError):
            api_call('unknown_end_point', 'client_id', 'user_id')

    @mock.patch('requests.get')
    def test_correct_url_was_called(self, mock_get):
        mock_get.return_value.status_code = 200
        api_call('end_point', 'client_id', 'user_id')
        mock_get.assert_called_once_with(
            'https://api.quizlet.com/2.0/users/user_id/end_point',
            {'client_id': 'client_id'}
        )

    @mock.patch('requests.get')
    def test_correct_output_was_returned(self, mock_get):
        response = mock.Mock()
        response.status_code = 200
        mock_get.return_value = response
        data = api_call('end_point', 'client_id', 'user_id')
        self.assertEqual(data, response.json())



"""User-available utilities."""
# ToDo: rename the module

from itertools import combinations
from typing import List, Tuple, Set

from quizler.lib import api_call
from quizler.models import WordSet


def get_user_sets(*api_envs):
    """Find all user sets."""
    data = api_call('sets', *api_envs)
    return [WordSet(wordset) for wordset in data]


def print_user_sets(wordsets: List[WordSet]):
    """Print all user sets by title and id."""
    if not wordsets:
        print('No sets found')
    else:
        print(f'Found sets: {len(wordsets)}')
        for wordset in wordsets:
            print(f'    {wordset}')


def get_common_terms(*api_envs) -> List[Tuple[str, str, Set[str]]]:
    """Get all term duplicates across all user word sets."""
    common_terms = []
    wordsets = get_user_sets(*api_envs)

    for wordset1, wordset2 in combinations(wordsets, 2):
        common = wordset1.has_common(wordset2)
        if common:
            common_terms.append((wordset1.title, wordset2.title, common))
    return common_terms


def print_common_terms(common_terms: List[Tuple[str, str, Set[str]]]):
    """Print common terms for each pair of word sets."""
    if not common_terms:
        print('No duplicates')
    else:
        for set_pair in common_terms:
            set1, set2, terms = set_pair
            print(f'{set1} and {set2} have in common:')
            for term in terms:
                print(f'    {term}')


def apply_regex(pattern, repl, set_name, *api_envs):
    """Apply regex replace to all terms in word set."""
    raise NotImplementedError  # ToDo: complete the utility



"""Additional none-user-visible utilities."""

import os

import requests


def get_api_envs():
    """Get required API keys from environment variables."""
    client_id = os.environ.get('CLIENT_ID')
    user_id = os.environ.get('USER_ID')
    if not client_id or not user_id:
        raise ValueError('API keys are not found in the environment')
    return client_id, user_id


def api_call(end_point, client_id, user_id):
    """Call given API end_point with API keys."""
    url = f'https://api.quizlet.com/2.0/users/{user_id}/{end_point}'
    params = {'client_id': client_id}
    response = requests.get(url, params)
    if response.status_code != 200:
        raise ValueError(
            f'Unknown end point, server returns {response.status_code}'
        )
    else:
        return response.json()



import unittest

from quizler.models import WordSet, Term


class TestTerm(unittest.TestCase):

    def test_correct_init(self):
        raw_data = {
            'definition': 'term definition',
            'id': 12345,
            'image': None,
            'rank': 0,
            'term': 'term'
        }
        term = Term(raw_data)
        self.assertEqual(term.definition, 'term definition')
        self.assertEqual(term.id, 12345)
        self.assertEqual(term.image, None)
        self.assertEqual(term.rank, 0)
        self.assertEqual(term.term, 'term')

    def test_wrong_term_json_structure(self):
        with self.assertRaises(ValueError):
            Term({'some data': '142% unexpected'})


class TestWordSet(unittest.TestCase):

    def setUp(self):
        self.word_set_1 = WordSet({
            'id': 0,
            'title': 'WordSet1',
            'terms': [{'term': 'term1'},
                      {'term': 'term2'}]
        })
        self.word_set_2 = WordSet({
            'id': 1,
            'title': 'WordSet2',
            'terms': [{'term': 'term2'},
                      {'term': 'term3'}]
        })

    def test_terms_set(self):
        self.assertSetEqual(self.word_set_1.term_set, {'term1', 'term2'})
        self.assertSetEqual(self.word_set_2.term_set, {'term2', 'term3'})

    def test_has_common(self):
        self.assertSetEqual(
            self.word_set_1.has_common(self.word_set_2),
            {'term2'}
        )

    def test_has_common_wrong_type(self):
        with self.assertRaises(ValueError):
            self.word_set_1.has_common([1, 2, 3])

    def test_repr(self):
        self.assertEqual(repr(self.word_set_1), 'WordSet1')
        self.assertEqual(repr(self.word_set_2), 'WordSet2')

    def test_str(self):
        self.assertEqual(str(self.word_set_1), 'WordSet1')
        self.assertEqual(str(self.word_set_2), 'WordSet2')



"""Entry point and CLI."""

import argparse

from quizler import get_api_envs, get_common_terms, apply_regex, get_user_sets
from quizler.utils import print_common_terms, print_user_sets


def create_parser():
    parser = argparse.ArgumentParser(description='Quizlet Utils')
    commands = parser.add_subparsers(title='Commands', dest='command')
    commands.required = True

    # Common
    commands.add_parser('common', help='Find duplicate terms across all sets')

    # Sets
    commands.add_parser('sets', help='List all your sets')

    # Apply
    apply = commands.add_parser('apply', help='Apply regex replace to set')
    apply.add_argument('pattern', type=str, help='Pattern to search for')
    apply.add_argument('repl', type=str, help='Replacement for pattern')
    apply.add_argument('set_name', type=str, help='Word set to apply to')

    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    api_envs = get_api_envs()
    if args.command == 'common':
        common_terms = get_common_terms(*api_envs)
        print_common_terms(common_terms)
    elif args.command == 'sets':
        user_sets = get_user_sets(*api_envs)
        print_user_sets(user_sets)
    elif args.command == 'apply':
        apply_regex(args.pattern, args.repl, args.set_name, *api_envs)


if __name__ == '__main__':
    main()



from distutils.core import setup

from setuptools import find_packages

setup(
    name='quizler',
    packages=find_packages(exclude=['tests', '*.test', '*.test.*']),
    version='0.0.3',
    description='Set of utils for Quizlet API',
    author='Pavel Karateev',
    author_email='karateev.pavel@ya.ru',
    url='https://github.com/lancelote/quizler',
    download_url='https://github.com/lancelote/quizler/archive/0.0.1.tar.gz',
    keywords=['quizlet', 'api'],
    entry_points={
        'console_scripts': [
            'quizler = quizler.main:main'
        ]
    },
    install_requires=['requests'],
)



"""OOP models for Quizlet terms abstractions."""
from typing import List, Dict


# ToDo: base class for Term and WordSet
class Term:
    """Quizlet term abstraction."""

    def __init__(self, raw_data):
        try:
            # ToDo: type annotations
            self.definition = raw_data['definition']
            self.id = raw_data['id']
            self.image = raw_data['image']
            self.rank = raw_data['rank']
            self.term = raw_data['term']
        except KeyError:
            raise ValueError('Unexpected term json structure')


class WordSet:
    """Quizlet set of terms and descriptions abstraction."""

    def __init__(self, raw_data):
        try:
            self.set_id: int = raw_data['id']
            self.title: str = raw_data['title']
            # ToDo: separate abstraction for Terms
            self.terms: List[Dict] = raw_data['terms']
        except KeyError:
            raise ValueError('Unexpected set json structure')

    def has_common(self, other):
        """Return set of common words between two word sets."""
        if not isinstance(other, WordSet):
            raise ValueError('Can compare only WordSets')
        return self.term_set & other.term_set

    @property
    def term_set(self):
        """Set of all terms in WordSet."""
        return {term['term'] for term in self.terms}

    def __eq__(self, other):
        # ToDo: introduce better solution
        return all((
            self.set_id == other.set_id,
            self.title == other.title,
            self.terms == other.terms
        ))

    def __str__(self):
        return f'{self.title}'

    __repr__ = __str__



import unittest
from unittest import mock

from quizler.models import WordSet
from quizler.utils import print_common_terms, get_common_terms, get_user_sets, \
    print_user_sets
from tests.utils import MockStdoutTestCase


@mock.patch('quizler.utils.get_user_sets')
class TestGetCommonTerms(unittest.TestCase):

    def test_one_common_term(self, mock_get_user_sets):
        mock_data = [
            WordSet({'id': 0,
                     'title': 'wordset1',
                     'terms': [{'term': 'term1'}, {'term': 'term2'}]}),
            WordSet({'id': 1,
                     'title': 'wordset2',
                     'terms': [{'term': 'term2'}, {'term': 'term3'}]}),
        ]
        mock_get_user_sets.return_value = mock_data
        self.assertEqual(
            get_common_terms(),
            [('wordset1', 'wordset2', {'term2'})]
        )

    def test_no_common_terms(self, mock_get_user_sets):
        mock_data = [
            WordSet({'id': 0,
                     'title': 'wordset1',
                     'terms': [{'term': 'term1'}, {'term': 'term2'}]}),
            WordSet({'id': 1,
                     'title': 'wordset2',
                     'terms': [{'term': 'term3'}, {'term': 'term4'}]})
        ]
        mock_get_user_sets.return_value = mock_data
        self.assertEqual(get_common_terms(), [])


class TestPrintCommonTerms(MockStdoutTestCase):

    def test_no_duplicates(self):
        print_common_terms([])
        self.assertStdout('No duplicates')

    def test_duplicates_in_one_pair(self):
        print_common_terms([('wordset1', 'wordset2', {'term1'})])
        self.assertStdout('wordset1 and wordset2 have in common:\n'
                          '    term1')

    def test_duplicates_in_multiple_pairs(self):
        print_common_terms([
            ('wordset1', 'wordset2', {'term12'}),
            ('wordset2', 'wordset3', {'term23'})
        ])
        self.assertStdout('wordset1 and wordset2 have in common:\n'
                          '    term12\n'
                          'wordset2 and wordset3 have in common:\n'
                          '    term23')


@mock.patch('quizler.utils.api_call')
class TestGetUserSets(unittest.TestCase):

    def test_there_are_sets(self, mock_api_call):
        mock_data = [
            {
                'id': 0,
                'title': 'wordset1',
                'terms': [{'term': 'term1'}, {'term': 'term2'}]
            },
            {
                'id': 1,
                'title': 'wordset2',
                'terms': [{'term': 'term3'}, {'term': 'term4'}]
            }
        ]
        # ToDo: bad testing - same logic in the code
        wordsets = [WordSet(wordset) for wordset in mock_data]
        mock_api_call.return_value = mock_data
        self.assertEqual(get_user_sets(), wordsets)

    def test_there_are_no_sets(self, mock_api_call):
        mock_api_call.return_value = []
        self.assertEqual(get_user_sets(), [])


class TestPrintUserSets(MockStdoutTestCase):

    def test_no_sets(self):
        print_user_sets([])
        self.assertStdout('No sets found')

    def test_one_set(self):
        print_user_sets([WordSet({'id': 0, 'title': 'wordset0', 'terms': []})])
        self.assertStdout('Found sets: 1\n'
                          '    wordset0')

    def test_two_sets(self):
        print_user_sets([
            WordSet({'id': 0, 'title': 'wordset0', 'terms': []}),
            WordSet({'id': 1, 'title': 'wordset1', 'terms': []}),
        ])
        self.assertStdout('Found sets: 2\n'
                          '    wordset0\n'
                          '    wordset1')


class TestApplyRegex(unittest.TestCase):
    ...
